/*
 * Image plugin to VDR (C++)
 *
 * (C) 2003 Kai Tobias Burwieck <kai@burwieck.net>
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <wait.h>
#include <math.h>
#include <locale.h>
/* NEU */
#include <linux/dvb/video.h>
#include <vdr/device.h>
#include <vdr/status.h>
#include <vdr/tools.h>

#include "player-image.h"
#include "setup-image.h"
#include "data-image.h"



const char *ImageCmd = "convert.sh";
const char *ImageCmdZoom = "convert_zoom.sh";
const char *ImageCmdJump = "convert_jump.sh";


//----------cImagePlayer-------------



class cImagePlayer:public cPlayer {
  private:
    bool active, sstarted, started;
    int inpipe[2], outpipe[2];
    cSlideShow *slideshow;
    cImage *current;

    const cImagePlayerControl *m_ipControl;

    bool Fork(void);
    void ClosePipe(void);
    void ImageControl(const char *format, ...);
  protected:
     cOsdBase * dvbOsd;
    virtual void Activate(bool On);
  public:
    const char *filename;
    char *change;
    int max, num, pid;
    bool cstarted;

     cImagePlayer(cSlideShow * sshow, const char *FileName);
     virtual ~ cImagePlayer();
    bool Active(void) {	return active;    } 
	void Play(void);
    void NextImage(int Step);
    void PrevImage(int Step);
    void GotoImage(int Pict);
    int ImageCount(void);
    void IncSlideTime(void);
    void DecSlideTime(void);
    void ChangeImage(char *change);
    bool Convert(const char *filename, char *c);
    bool ConvertZoom(const char *filename, char *change, int xRes, int yRes,
		     int xPos, int yPos);
    bool ConvertJump(int Step);
    bool ShowImage(void);
    void send_pes_packet(unsigned char *data, int len, int timestamp);

  /** Returns the current and total frame index, optionally snapped to the
      nearest I-frame.*/
    virtual bool GetIndex(int &Current, int &Total, bool SnapToIFrame);

    void setController(cImagePlayerControl * ipControl) {
	m_ipControl = ipControl;
    }
};

bool cImagePlayer::GetIndex(int &Current, int &Total, bool SnapToIFrame)
{
    // Notify other status-plugins with one picture as one second
    Current = num * FRAMESPERSEC * ImageSetup.SSsec; 
    // Notify other status-plugins count of picture as totalcount of seconds
    Total = max * FRAMESPERSEC * ImageSetup.SSsec;
	
    return true;
}

cImagePlayer::cImagePlayer(cSlideShow * sshow, const char *FileName):cPlayer()
{
    m_ipControl = 0;

    num = 0;
    max = 0;
    inpipe[0] = inpipe[1] = -1;
    outpipe[0] = outpipe[1] = -1;
    pid = -1;
    slideshow = sshow;
    if(slideshow)
	{
	    filename = slideshow->First()->Name();
	    max = slideshow->Count();
	    current = slideshow->First();
	    num++;
	}
    active = true;
    started = false;
    sstarted = false;
}



cImagePlayer::~cImagePlayer()
{
    kill(pid, SIGTERM);
    waitpid(pid, 0, 0);
    ClosePipe();
    Detach();
}



void cImagePlayer::Activate(bool On)
{
    if(On)
	{
	    if(slideshow && !started)
		{
		    Convert(filename, "");
		}
	}
}



void cImagePlayer::ClosePipe(void)
{
    if(inpipe[0] >= 0)
	{
	    close(inpipe[0]);
	    inpipe[0] = -1;
	}
    if(inpipe[1] >= 0)
	{
	    close(inpipe[1]);
	    inpipe[1] = -1;
	}
    if(outpipe[0] >= 0)
	{
	    close(outpipe[0]);
	    outpipe[0] = -1;
	}
    if(outpipe[1] >= 0)
	{
	    close(outpipe[1]);
	    outpipe[1] = -1;
	}
}



void cImagePlayer::Play(void)
{
    if(current)
	{
	    if(!cstarted)
		{
		    ShowImage();
		}
	}
}



void cImagePlayer::NextImage(int Step)
{
    if(current && num < max)
	{
	    cImage *newcurr;
	    for (int i = 0; (i < Step) && (num < max); i++)
		{
		    newcurr = slideshow->cList < cImage >::Next(current);
		    current = newcurr;
		    num++;
		}

	    filename = current->Name();
	    Convert(filename, "");
    } else if(ImageSetup.AutoRepeat)
	{
	    cImage *newcurr;

	    newcurr = slideshow->cList < cImage >::First();
	    current = newcurr;
	    num = 1;

	    filename = current->Name();

	    Convert(filename, "");
	}
}



void cImagePlayer::PrevImage(int Step)
{
    if(current && num > 1)
	{
	    cImage *newcurr;

	    for (int i = 0; (i < Step) && (num > 1); i++)
		{
		    newcurr = slideshow->cList < cImage >::Prev(current);
		    current = newcurr;
		    num--;
		}

	    filename = current->Name();
	    Convert(filename, "");

    } else if(ImageSetup.AutoRepeat)
	{
	    cImage *newcurr;

	    newcurr = slideshow->cList < cImage >::Last();
	    current = newcurr;
	    num = max;

	    filename = current->Name();

	    Convert(filename, "");
	}
}


void cImagePlayer::GotoImage(int Pict)
{
    if(current && Pict > 0 && Pict < max)
	{
	    cImage *newcurr;
	    newcurr = slideshow->cList < cImage >::First();
	    current = newcurr;
	    num = 0;
	    for (int i = 0; i < Pict; i++)
		{
		    newcurr = slideshow->cList < cImage >::Next(current);
		    current = newcurr;
		    num++;
		}
	    filename = current->Name();
	    Convert(filename, "");
	}
}


int cImagePlayer::ImageCount(void)
{
    return max;
}

void cImagePlayer::IncSlideTime(void)
{
    if(ImageSetup.SSsec < 120)
	{
	    ImageSetup.SSsec++;
	}
}

void cImagePlayer::DecSlideTime(void)
{
    if(ImageSetup.SSsec > 2)
	{
	    ImageSetup.SSsec--;
	}
}

void cImagePlayer::ChangeImage(char *change)
{
    if(change)
	{
	    if(current)
		{
		    filename = current->Name();
		    if(change)
			Convert(filename, change);
		}
	}
}

bool cImagePlayer::Convert(const char *filename, char *change)
{
    if(m_ipControl)
		cStatus::MsgReplaying(m_ipControl, filename);

    cstarted = true;
    if(inpipe[0] >= 0)
	{
	    close(inpipe[0]);
	    inpipe[0] = -1;
	}
    if(inpipe[1] >= 0)
	{
	    close(inpipe[1]);
	    inpipe[1] = -1;
	}
    if(outpipe[0] >= 0)
	{
	    close(outpipe[0]);
	    outpipe[0] = -1;
	}
    if(outpipe[1] >= 0)
	{
	    close(outpipe[1]);
	    outpipe[1] = -1;
	}
    pid = fork();
    if(pid == -1)
	{
	    esyslog("image: convert failed: (%d) %s", errno, strerror(errno));
	    return false;
	}
    if(pid == 0)
	{
	    dsyslog("image: script startet (pid=%d)", getpid());
	    char cmd[64 + PATH_MAX * 2];
	    snprintf(cmd, sizeof(cmd), "%s \"%s\" \"%s\" %s", ImageCmd,
		     filename, ImageSetup.TempDir, change ? change : "");
	    execle("/bin/sh", "sh", "-c", cmd, 0, environ);
	    esyslog("image: exec failed for %s: (%d) %s", cmd, errno,
		    strerror(errno));
	    exit(127);
	}
    return true;
}

bool cImagePlayer::ConvertZoom(const char *filename, char *change, int xRes,
			       int yRes, int xPos, int yPos)
{
    cstarted = true;
    if(inpipe[0] >= 0)
	{
	    close(inpipe[0]);
	    inpipe[0] = -1;
	}
    if(inpipe[1] >= 0)
	{
	    close(inpipe[1]);
	    inpipe[1] = -1;
	}
    if(outpipe[0] >= 0)
	{
	    close(outpipe[0]);
	    outpipe[0] = -1;
	}
    if(outpipe[1] >= 0)
	{
	    close(outpipe[1]);
	    outpipe[1] = -1;
	}
    pid = fork();
    if(pid == -1)
	{
	    esyslog("image: convert failed: (%d) %s", errno, strerror(errno));
	    return false;
	}
    if(pid == 0)
	{
	    dsyslog("image: script startet (pid=%d)", getpid());
	    static char cmd[128 + PATH_MAX * 2];
	    snprintf(cmd, sizeof(cmd), "%s \"%s\" \"%s\" %s %d %d %d %d",
		     ImageCmdZoom, filename, ImageSetup.TempDir,
		     change ? change : "original", xRes, yRes, xPos, yPos);
	    dsyslog("image: %s", cmd);
	    execle("/bin/sh", "sh", "-c", cmd, 0, environ);
	    esyslog("image: exec failed for %s: (%d) %s", cmd, errno,
		    strerror(errno));
	    exit(127);
	}
    return true;
}

bool cImagePlayer::ConvertJump(int Step)
{
#define MAX_BILDER	9
    static char FileName[MAX_BILDER][PATH_MAX];
    int i;
    cImage *newcurr;
    int Bilder = 0;

    cstarted = true;

    for (i = 0; i < MAX_BILDER; i++)
	memset(FileName[i], 0, PATH_MAX);

    if(current && num > 0)
	{
	    // Nun denn: zuerst springen wir mal ein bisschen
	    if(Step < 0)
		{
		    Step = -Step;
		    for (i = 0; (i < Step) && (num > 1); i++)
			{
			    newcurr =
				slideshow->cList < cImage >::Prev(current);
			    current = newcurr;
			    num--;
			}

		    filename = current->Name();
	    } else if(Step > 0)
		{

		    for (i = 0; (i < Step) && (num < max); i++)
			{
			    newcurr =
				slideshow->cList < cImage >::Next(current);
			    current = newcurr;
			    num++;
			}
		    filename = current->Name();
		}

	    if((max - num) < MAX_BILDER)
		{		// es sollten schon 9 Bilder angezeigt werden
		    for (; ((max - num) < MAX_BILDER) && (num > 1);)
			{
			    newcurr =
				slideshow->cList < cImage >::Prev(current);
			    current = newcurr;
			    num--;
			}
		    filename = current->Name();
		}
	    // um danach maximal MAX_BILDER Dateinamen zu holen
	    newcurr = current;
	    for (i = 0; (i < MAX_BILDER) && ((num + i) < max); i++)
		{
		    strncpy(FileName[i], newcurr->Name(), PATH_MAX);
		    dsyslog("image: File%d: %s", i, newcurr->Name());
		    newcurr = slideshow->cList < cImage >::Next(newcurr);
		    Bilder++;
		}
	}

    if(inpipe[0] >= 0)
	{
	    close(inpipe[0]);
	    inpipe[0] = -1;
	}

    if(inpipe[1] >= 0)
	{
	    close(inpipe[1]);
	    inpipe[1] = -1;
	}

    if(outpipe[0] >= 0)
	{
	    close(outpipe[0]);
	    outpipe[0] = -1;
	}

    if(outpipe[1] >= 0)
	{
	    close(outpipe[1]);
	    outpipe[1] = -1;
	}

    pid = fork();

    if(pid == -1)
	{

	    esyslog("image: convert failed: (%d) %s", errno, strerror(errno));

	    return false;

	}

    if(pid == 0)
	{
	    static char cmd[PATH_MAX * (MAX_BILDER + 3)];

	    dsyslog("image: script startet (pid=%d)", getpid());

	    snprintf(cmd, sizeof(cmd),
		     "%s \"%s\" %s %s \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%s\"",
		     ImageCmdJump, ImageSetup.TempDir,
		     (Bilder > 6) ? "3x3" : ((Bilder > 3) ? "3x2" : "3x1"),
		     "256x192", FileName[0], FileName[1], FileName[2],
		     FileName[3], FileName[4], FileName[5], FileName[6],
		     FileName[7], FileName[8]);
	    dsyslog("%s", cmd);
	    execle("/bin/sh", "sh", "-c", cmd, 0, environ);
	    esyslog("image: exec failed for %s: (%d) %s", cmd, errno,
		    strerror(errno));
	    exit(127);
	}
    return true;
}



bool cImagePlayer::ShowImage(void)
{
    if(!sstarted)
	{
	    uchar *buffer;
	    int filefd;
	    ssize_t nRead;

	    struct stat st;
	    char tempfile[138];

	    struct video_still_picture sp;
	    snprintf(tempfile, sizeof(tempfile), "%s%s", ImageSetup.TempDir,
		     "/image.mpg");
	    if((filefd = open(tempfile, O_RDONLY)) >= 0)
		{
		    dsyslog("file %s opened", tempfile);
		    fstat(filefd, &st);

		    sp.iFrame = (char *)malloc(st.st_size);
		    sp.size = st.st_size;

		    if(sp.iFrame)
			{
			    nRead = read(filefd, sp.iFrame, sp.size);
			    dsyslog("image: %d bytes fread from %s", nRead, tempfile);

			    buffer = (uchar *) sp.iFrame;

			    if(ImageSetup.UseDeviceStillPicture)
				{
				    DeviceStillPicture(buffer, sp.size);
			    } else
				{
				    for (int i = 1; i <= 25; i++)
					send_pes_packet(buffer, sp.size, i);
				}

			    free(sp.iFrame);
		    } else
			{
			    esyslog("image: alloc of %d bytes failed",sp.size);
			}
		    close(filefd);
	    } else
		{
		    esyslog("image: read file %s failed (%d) %s", tempfile ,errno, strerror(errno));
		}
	    sstarted = false;
	}
    else
	{
	    sstarted = true;
	}

    return true;
}



void cImagePlayer::send_pes_packet(unsigned char *data, int len, int timestamp)
{
#define PES_MAX_SIZE 2048
    int ptslen = timestamp ? 5 : 1;
    static unsigned char pes_header[PES_MAX_SIZE];

    pes_header[0] = pes_header[1] = 0;
    pes_header[2] = 1;
    pes_header[3] = 0xe0;

    while(len > 0)
	{
	    int payload_size = len;
	    if(6 + ptslen + payload_size > PES_MAX_SIZE)
		payload_size = PES_MAX_SIZE - (6 + ptslen);

	    pes_header[4] = (ptslen + payload_size) >> 8;
	    pes_header[5] = (ptslen + payload_size) & 255;

	    if(ptslen == 5)
		{
		    int x;
		    x = (0x02 << 4) | (((timestamp >> 30) & 0x07) << 1) | 1;
		    pes_header[8] = x;
		    x = ((((timestamp >> 15) & 0x7fff) << 1) | 1);
		    pes_header[7] = x >> 8;
		    pes_header[8] = x & 255;
		    x = ((((timestamp) & 0x7fff) < 1) | 1);
		    pes_header[9] = x >> 8;
		    pes_header[10] = x & 255;
	    } else
		{
		    pes_header[6] = 0x0f;
		}

	    memcpy(&pes_header[6 + ptslen], data, payload_size);
	    PlayVideo(pes_header, 6 + ptslen + payload_size);

	    len -= payload_size;
	    data += payload_size;
	    ptslen = 1;
	}
}





//----------cImagePlayerControl---------------



cImagePlayerControl::cImagePlayerControl(cSlideShow * sshow)
	:cControl(player = new cImagePlayer(sshow, FileName = sshow->First()-> Name()))
{
	// Remember this object on my player object
    player->setController(this);
	// Notity all cStatusMonitor
    cStatus::MsgReplaying(this, "Image");

    startime = time(NULL);
}



cImagePlayerControl::~cImagePlayerControl()
{
	// Notity cleanup all cStatusMonitor
    cStatus::MsgReplaying(this, NULL);
    Stop();
}



bool cImagePlayerControl::Active(void)
{
    if(ss)
	{
	    t2 = time(NULL);
	    t3 = t2 - startime;
	    if(t3 >= ImageSetup.SSsec)
		{
		    if( /*player->num < player->max && */ !player->cstarted)
			{
			    startime = time(NULL);
			    NextImage(1);
			}
		}
	}
    if(player->cstarted)
	{
	    startime = time(NULL);
	    waitpid(player->pid, 0, WNOHANG);
	    if(kill(player->pid, 0) < 0 && errno == ESRCH)
		{
		    player->cstarted = false;
		    player->Play();
		}
	}
    cstarted = player->cstarted;
    filename = player->filename;
    aktnum = player->num;
    maxnum = player->max;
    return player && player->Active();
}

void cImagePlayerControl::Stop(void)
{
    delete player;
    player = 0;
}

void cImagePlayerControl::Pause(void)
{
    // for Slideshow
}



void cImagePlayerControl::Play(void)
{
    // for SlideShow
}

void cImagePlayerControl::NextImage(int Step)
{
    if(player)
		player->NextImage(Step);
}



void cImagePlayerControl::PrevImage(int Step)
{
    if(player)
		player->PrevImage(Step);
}


void cImagePlayerControl::GotoImage(int Pict)
{
    if(player)
		player->GotoImage(Pict);
}


int cImagePlayerControl::ImageCount(void)
{
    if(player)
		return player->ImageCount();
    return 0;

}


void cImagePlayerControl::RFlipImage(void)
{
    if(player)
		player->ChangeImage("Right");
}



void cImagePlayerControl::LFlipImage(void)
{
    if(player)
		player->ChangeImage("Left");
}



void cImagePlayerControl::OriginalImage(void)
{
    if(player)
		player->ChangeImage("Original");
}



void cImagePlayerControl::Jump(void)
{
	
}


void cImagePlayerControl::IncSlideTime(void)
{
    if(player)
		player->IncSlideTime();
}


void cImagePlayerControl::DecSlideTime(void)
{
    if(player)
		player->DecSlideTime();
}

bool cImagePlayerControl::ConvertZoom(const char *filename, char *change,
				      int xRes, int yRes, int xPos, int yPos)
{
    if(player)
		return player->ConvertZoom(filename, change, xRes, yRes, xPos, yPos);
    return 0;
}

bool cImagePlayerControl::ConvertJump(int Step)
{
    if(player)
		return player->ConvertJump(Step);
    return 0;
}
